#include <stdio.h>
#include <string.h>
#include "lcd.h"

static void make_pins_output(struct pin pins[NUM_PINS]);
static void unexport_pins(struct pin pins[NUM_PINS]);
static void set_pin_value(struct pin pin, int value);
static void set_pin_value_multiple(struct pin pins[NUM_PINS], int values[NUM_PINS], int clock);
static void exec_command(struct pin pins[NUM_PINS], int command[NUM_PINS]);
static void print_char(struct pin pins[NUM_PINS], char c);
static void move_to_line(struct pin pins[NUM_PINS], int line);

// will allocate memory for the screen. Use screen_deinit to free
// number of pins MUST be 6
void screen_init(struct pin pins[NUM_PINS]) {
    make_pins_output(pins);

    // pins are E RS D4 D5 D6 D7
    int commands[6][6] = {
        {0, 0, 1, 1, 0, 0},
        {0, 0, 0, 1, 0, 0}, // 4-bit mode
        {0, 0, 0, 0, 0, 0}, // enable display pt1
        {0, 0, 0, 0, 1, 1}, // enable display pt2
        {0, 0, 0, 1, 0, 0}, // two-line mode pt1
        {0, 0, 0, 0, 0, 1}  // two-line mode pt2
    };

    for (int i = 0; i < 6; i++) {
        exec_command(pins, commands[i]);
    }
}

// free memory used by screen
void screen_deinit(struct pin pins[NUM_PINS]) {
    // set screen back to 8 bit mode
    int command[] = {0, 0, 1, 1, 0, 0};
    exec_command(pins, command);

    unexport_pins(pins);
}

void screen_clear(struct pin pins[NUM_PINS]) {
    int commands[][6] = {
        {0, 0, 0, 0, 0, 0},
        {0, 0, 1, 0, 0, 0}
    };

    for (int i = 0; i < 2; i++) {
        exec_command(pins, commands[i]);
    }
}

void screen_print(struct pin pins[NUM_PINS], char *str, int len, int line) {
    move_to_line(pins, line);
    for (int i = 0; i < len; i++) {
        print_char(pins, str[i]);
    }

    // so we don't end up random characters by overwriting the line
    for (int i = len; i < 20; i++) {
        print_char(pins, ' ');
    }
}

static void move_to_line(struct pin pins[NUM_PINS], int line) {
    // like chars, this needs to be split in 2 parts
    int line_commands1[2][6] = {
        {0, 0, 0, 0, 0, 1},
        {0, 0, 0, 0, 1, 1}
    };

    int line_commands2[2][6] = {
        {0, 0, 0, 0, 0, 0},
        {0, 0, 0, 0, 0, 0}
    };

    switch (line) {
        case 0:
            exec_command(pins, line_commands1[line]);
            exec_command(pins, line_commands2[line]);
            break;
        default:
            exec_command(pins, line_commands1[1]);
            exec_command(pins, line_commands2[1]);
            break;
    }
}

static void print_char(struct pin pins[NUM_PINS], char c) {
    // printing characters on 4-bit is split across 2 commands
    // RS is set to 1 because we're loading stuff into the register now
    int command1[NUM_PINS] = {0, 1};
    int command2[NUM_PINS] = {0, 1};

    // the commands below start at 2 because that's
    // the index of first data pin

    // this is all generated code
    switch (c) {
        case 'a':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,0,0}, sizeof(int[4]));
            break;
        case 'b':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            break;
        case 'c':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            break;
        case 'd':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            break;
        case 'e':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            break;
        case 'f':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            break;
        case 'g':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            break;
        case 'h':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,0,1}, sizeof(int[4]));
            break;
        case 'i':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,0,1}, sizeof(int[4]));
            break;
        case 'j':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,0,1}, sizeof(int[4]));
            break;
        case 'k':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,0,1}, sizeof(int[4]));
            break;
        case 'l':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,1,1}, sizeof(int[4]));
            break;
        case 'm':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,1,1}, sizeof(int[4]));
            break;
        case 'n':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,1,1}, sizeof(int[4]));
            break;
        case 'o':
            memcpy(command1+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,1,1}, sizeof(int[4]));
            break;
        case 'p':
            memcpy(command1+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,0,0}, sizeof(int[4]));
            break;
        case 'q':
            memcpy(command1+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,0,0}, sizeof(int[4]));
            break;
        case 'r':
            memcpy(command1+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            break;
        case 's':
            memcpy(command1+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            break;
        case 't':
            memcpy(command1+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            break;
        case 'u':
            memcpy(command1+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            break;
        case 'v':
            memcpy(command1+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            break;
        case 'w':
            memcpy(command1+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            break;
        case 'x':
            memcpy(command1+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,0,1}, sizeof(int[4]));
            break;
        case 'y':
            memcpy(command1+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,0,1}, sizeof(int[4]));
            break;
        case 'z':
            memcpy(command1+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,0,1}, sizeof(int[4]));
            break;
        case 'A':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,0,0}, sizeof(int[4]));
            break;
        case 'B':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            break;
        case 'C':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            break;
        case 'D':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            break;
        case 'E':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            break;
        case 'F':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            break;
        case 'G':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            break;
        case 'H':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,0,1}, sizeof(int[4]));
            break;
        case 'I':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,0,1}, sizeof(int[4]));
            break;
        case 'J':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,0,1}, sizeof(int[4]));
            break;
        case 'K':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,0,1}, sizeof(int[4]));
            break;
        case 'L':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,1,1}, sizeof(int[4]));
            break;
        case 'M':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,1,1}, sizeof(int[4]));
            break;
        case 'N':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,1,1}, sizeof(int[4]));
            break;
        case 'O':
            memcpy(command1+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,1,1}, sizeof(int[4]));
            break;
        case 'P':
            memcpy(command1+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,0,0}, sizeof(int[4]));
            break;
        case 'Q':
            memcpy(command1+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,0,0}, sizeof(int[4]));
            break;
        case 'R':
            memcpy(command1+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            break;
        case 'S':
            memcpy(command1+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            break;
        case 'T':
            memcpy(command1+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            break;
        case 'U':
            memcpy(command1+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            break;
        case 'V':
            memcpy(command1+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            break;
        case 'W':
            memcpy(command1+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            break;
        case 'Y':
            memcpy(command1+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,0,1}, sizeof(int[4]));
            break;
        case 'Z':
            memcpy(command1+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,0,1}, sizeof(int[4]));
            break;
        case '1':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,0,0}, sizeof(int[4]));
            break;
        case '2':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            break;
        case '3':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            break;
        case '4':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            break;
        case '5':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            break;
        case '6':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            break;
        case '7':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,1,0}, sizeof(int[4]));
            break;
        case '8':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,0,1}, sizeof(int[4]));
            break;
        case '9':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,0,1}, sizeof(int[4]));
            break;
        case '0':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,0,0}, sizeof(int[4]));
            break;
        case ':':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            break;
        case ';':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,0,1}, sizeof(int[4]));
            break;
        case '<':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,1,1}, sizeof(int[4]));
            break;
        case '=':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,1,1}, sizeof(int[4]));
            break;
        case '>':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,1,1}, sizeof(int[4]));
            break;
        case '?':
            memcpy(command1+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,1,1}, sizeof(int[4]));
            break;
        case '!':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,0,0}, sizeof(int[4]));
            break;
        case '#':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,0,0}, sizeof(int[4]));
            break;
        case '$':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,1,0}, sizeof(int[4]));
            break;
        case '%':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,1,0}, sizeof(int[4]));
            break;
        case '&':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,1,0}, sizeof(int[4]));
            break;
        case '(':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,0,1}, sizeof(int[4]));
            break;
        case ')':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,0,1}, sizeof(int[4]));
            break;
        case '*':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,0,1}, sizeof(int[4]));
            break;
        case '+':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,0,1}, sizeof(int[4]));
            break;
        case ',':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,0,1,1}, sizeof(int[4]));
            break;
        case '-':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,0,1,1}, sizeof(int[4]));
            break;
        case '.':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,1,1}, sizeof(int[4]));
            break;
        case '/':
            memcpy(command1+2, (int[4]){0,1,0,0}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){1,1,1,1}, sizeof(int[4]));
            break;
        case ' ':
            memcpy(command1+2, (int[4]){1,1,1,1}, sizeof(int[4]));
            memcpy(command2+2, (int[4]){0,1,1,1}, sizeof(int[4]));
            break;
        default:
            break;
    }

    exec_command(pins, command1);
    exec_command(pins, command2);
}

static void make_pins_output(struct pin pins[NUM_PINS]) {
    FILE *f;
    for (int i = 0; i < NUM_PINS; i++) {
        f = fopen("/sys/class/gpio/export", "w");
        if (f != NULL) {
            // TODO: what if f IS null?
            fprintf(f, "%d", pins[i].gpio_number);
            fclose(f);
        } else {
            printf("Unable to export pin\n");
        }

        char filename[64];
        sprintf(filename, "/sys/class/gpio/gpio%d/direction", pins[i].gpio_number);

        f = fopen(filename, "w");
        if (f != NULL) {
            fprintf(f, "out");
            fclose(f);
        } else {
            printf("Unable to set direction [%d]\n", pins[i].gpio_number);
            printf("%s\n", filename);
        }

        set_pin_value(pins[i], 0);
    }
};

static void set_pin_value(struct pin pin, int value) {
    char filename[64];
    sprintf(filename, "/sys/class/gpio/gpio%d/value", pin.gpio_number);

    FILE *f = fopen(filename, "w");
    if (f != NULL) {
        fprintf(f, "%d", value);
        fclose(f);
    } else {
        printf("unable to set pin value\n");
    }
}

static void set_pin_value_multiple(struct pin pins[NUM_PINS], int values[NUM_PINS], int clock) {
    set_pin_value(pins[0], clock);

    for (int i = 1; i < NUM_PINS; i++) {
        set_pin_value(pins[i], values[i]);
    }
}

static void exec_command(struct pin pins[NUM_PINS], int command[NUM_PINS]) {
    // you have to enable and disable the clock to make the instruction be read
    set_pin_value_multiple(pins, command, 0);
    set_pin_value_multiple(pins, command, 1);
    set_pin_value_multiple(pins, command, 0);
}

static void unexport_pins(struct pin pins[NUM_PINS]) {
    FILE *f;
    for (int i = 0; i < NUM_PINS; i++) {
        f = fopen("/sys/class/gpio/unexport", "w");
        if (f != NULL) {
            fprintf(f, "%d", pins[i].gpio_number);
            fclose(f);
        } else {
            printf("Unable to unexport pin\n");
        }
    }
}
